#include <ntddk.h>

#define CHILDDEVICECOUNT 2
#define BUSDRIVERDEVICE 0
#define CHILDDRIVERDEVICE 1
#define CHILDDEVICEDEVICEID L"HAJESOFT\\CHILD_DEVICEID"
#define CHILDDEVICEHARDWAREID L"HAJESOFT\\CHILD_HARDWAREID"
#define CHILDDEVICECOMPATIBLEID L"HAJESOFT\\CHILD_COMPATIBLEID"
#define CHILDDEVICEINSTANCEID1 L"0001"
#define CHILDDEVICEINSTANCEID2 L"0002"
#define CHILDDEVICETEXT  L"HAJESOFT SAMPLE Child Device"

typedef struct
{
	int DeviceObjectType;
	PDEVICE_OBJECT 	NextLayerDeviceObject;
	PDEVICE_OBJECT ChildDeviceObject[CHILDDEVICECOUNT];
	PDEVICE_OBJECT ParentDeviceObject;
}DEVICE_EXTENSION, *PDEVICE_EXTENSION;

NTSTATUS
BUSDRIVER_AddDevice
	(
	IN PDRIVER_OBJECT DriverObject,         
	IN PDEVICE_OBJECT PhysicalDeviceObject  
	)
{
	NTSTATUS returnStatus = STATUS_SUCCESS;         
	PDEVICE_OBJECT DeviceObject = NULL;             
	PDEVICE_EXTENSION deviceExtension;

	returnStatus = IoCreateDevice                                       
			(                                               
				DriverObject,                               
				sizeof(DEVICE_EXTENSION),                
				NULL,                   
				FILE_DEVICE_UNKNOWN,
				0,                                          
				FALSE,                                      
				&DeviceObject       
			);

	deviceExtension = DeviceObject->DeviceExtension;
	deviceExtension->DeviceObjectType = BUSDRIVERDEVICE;
	deviceExtension->NextLayerDeviceObject =
		IoAttachDeviceToDeviceStack (
					DeviceObject,
					PhysicalDeviceObject	
	);

	DeviceObject->Flags &= ~DO_DEVICE_INITIALIZING; 

	return returnStatus;
}

NTSTATUS
MyCompletionRoutine
	(
	IN PDEVICE_OBJECT DeviceObject,   
	IN PIRP Irp,
	IN PVOID Context
	)
{
	PKEVENT pEvent;
	pEvent = (PKEVENT)Context;
	KeSetEvent( pEvent, 0, FALSE );
	return STATUS_MORE_PROCESSING_REQUIRED;
}

NTSTATUS
BusDriverDevice_PnpDispatch
	(
	IN PDEVICE_OBJECT DeviceObject,   
	IN PIRP Irp                       
	)
{
	PIO_STACK_LOCATION pStack; 
	PDEVICE_EXTENSION deviceExtension; 
	PDEVICE_OBJECT NextLayerDeviceObject;
	NTSTATUS returnStatus;
	
	pStack = IoGetCurrentIrpStackLocation ( Irp );
	deviceExtension = DeviceObject->DeviceExtension;
	NextLayerDeviceObject = deviceExtension->NextLayerDeviceObject;

	switch ( pStack->MinorFunction )
	{
		case IRP_MN_START_DEVICE :
		{
			{
				KEVENT Event;
				PDEVICE_EXTENSION childDeviceExtension;

				KeInitializeEvent( &Event, NotificationEvent, FALSE );

				IoCopyCurrentIrpStackLocationToNext( Irp );
				IoSetCompletionRoutine( 
					Irp, 
					MyCompletionRoutine, 
					&Event, 
					TRUE, 
					TRUE, 
					TRUE );
				returnStatus = IoCallDriver( NextLayerDeviceObject, Irp );
				if( returnStatus == STATUS_PENDING )
				{
					KeWaitForSingleObject( 
						&Event, 
						Executive, 
						KernelMode, 
						FALSE, 
						NULL );
					returnStatus = Irp->IoStatus.Status;
				}
				if( NT_SUCCESS( returnStatus ) )
				{
					// New Child Device1 
					IoCreateDevice                                       
					(                                               
						DeviceObject->DriverObject,                               
						sizeof(DEVICE_EXTENSION),                
						NULL,                   
						FILE_DEVICE_UNKNOWN,
						0,                                          
						FALSE,                                      
						&deviceExtension->ChildDeviceObject[0]       
					);
					childDeviceExtension = (PDEVICE_EXTENSION)deviceExtension->ChildDeviceObject[0]->DeviceExtension;
					childDeviceExtension->DeviceObjectType = CHILDDRIVERDEVICE;
					childDeviceExtension->ParentDeviceObject = DeviceObject;

					// New Child Device2 
					IoCreateDevice                                       
					(                                               
						DeviceObject->DriverObject,                               
						sizeof(DEVICE_EXTENSION),                
						NULL,                   
						FILE_DEVICE_UNKNOWN,
						0,                                          
						FALSE,                                      
						&deviceExtension->ChildDeviceObject[1]       
					);
					childDeviceExtension = (PDEVICE_EXTENSION)deviceExtension->ChildDeviceObject[1]->DeviceExtension;
					childDeviceExtension->DeviceObjectType = CHILDDRIVERDEVICE;
					childDeviceExtension->ParentDeviceObject = DeviceObject;
				}

				IoCompleteRequest( Irp, IO_NO_INCREMENT );
				return returnStatus;
			}			
		}
		break;			

		case IRP_MN_REMOVE_DEVICE :
		{ 
			IoDetachDevice ( NextLayerDeviceObject );								
			IoDeleteDevice ( DeviceObject );									
			Irp->IoStatus.Status = STATUS_SUCCESS;
		}
		break;

		case IRP_MN_QUERY_DEVICE_RELATIONS :
		{ 
			{
				PDEVICE_RELATIONS pRelations;

				if( pStack->Parameters.QueryDeviceRelations.Type == BusRelations )
				{
					pRelations = (PDEVICE_RELATIONS)ExAllocatePool( 
						PagedPool, 
						sizeof(ULONG) + sizeof(PDEVICE_OBJECT) * (CHILDDEVICECOUNT + 1) );
					pRelations->Count = CHILDDEVICECOUNT;
					pRelations->Objects[0] = deviceExtension->ChildDeviceObject[0];
					pRelations->Objects[1] = deviceExtension->ChildDeviceObject[1];
					pRelations->Objects[2] = 0;

					ObReferenceObject( pRelations->Objects[0] );
					ObReferenceObject( pRelations->Objects[1] );

					Irp->IoStatus.Information = (ULONG)pRelations;				
					Irp->IoStatus.Status = STATUS_SUCCESS;
				}
			}
		}
		break;
	} 

	IoSkipCurrentIrpStackLocation( Irp ); 
	returnStatus = IoCallDriver(NextLayerDeviceObject, Irp);

	return returnStatus;
}

NTSTATUS
QueryDeviceCaps( IN  PIRP   Irp )
{
	
    PIO_STACK_LOCATION      pStack;
    PDEVICE_CAPABILITIES    deviceCapabilities;
    NTSTATUS                status;
    
    pStack = IoGetCurrentIrpStackLocation (Irp);
	
    deviceCapabilities = pStack->Parameters.DeviceCapabilities.Capabilities;
	
    if(deviceCapabilities->Version != 1 || 
		deviceCapabilities->Size < sizeof(DEVICE_CAPABILITIES))
    {
		return STATUS_UNSUCCESSFUL; 
    }
    
    deviceCapabilities->DeviceState[PowerSystemWorking] = PowerDeviceD0;
    deviceCapabilities->DeviceState[PowerSystemSleeping1] = PowerDeviceD3;
    deviceCapabilities->DeviceState[PowerSystemSleeping2] = PowerDeviceD3;
    deviceCapabilities->DeviceState[PowerSystemSleeping3] = PowerDeviceD3;
    deviceCapabilities->SystemWake = PowerSystemUnspecified;
    deviceCapabilities->DeviceWake = PowerDeviceUnspecified;
    deviceCapabilities->DeviceD1 = FALSE;
    deviceCapabilities->DeviceD2 = FALSE;
    deviceCapabilities->WakeFromD0 = FALSE;
    deviceCapabilities->WakeFromD1 = FALSE;
    deviceCapabilities->WakeFromD2 = FALSE;
    deviceCapabilities->WakeFromD3 = FALSE;
    deviceCapabilities->D1Latency = 0;
    deviceCapabilities->D2Latency = 0;
    deviceCapabilities->D3Latency = 0;
    deviceCapabilities->EjectSupported = FALSE;
    deviceCapabilities->HardwareDisabled = FALSE;
    deviceCapabilities->Removable = TRUE;
	deviceCapabilities->SurpriseRemovalOK = TRUE;
    deviceCapabilities->UniqueID = TRUE;
    deviceCapabilities->SilentInstall = FALSE;
    return STATUS_SUCCESS;
}

NTSTATUS
ChildDriverDevice_PnpDispatch
	(
	IN PDEVICE_OBJECT DeviceObject,   
	IN PIRP Irp                       
	)
{
	PIO_STACK_LOCATION pStack; 
	PDEVICE_EXTENSION deviceExtension; 
	PDEVICE_OBJECT NextLayerDeviceObject;
	NTSTATUS returnStatus = Irp->IoStatus.Status;
	
	pStack = IoGetCurrentIrpStackLocation ( Irp );
	deviceExtension = DeviceObject->DeviceExtension;

	switch ( pStack->MinorFunction )
	{
		case IRP_MN_QUERY_CAPABILITIES:
		{
			returnStatus = QueryDeviceCaps( Irp );
			break;
		}

		case IRP_MN_START_DEVICE :
		case IRP_MN_QUERY_STOP_DEVICE:
		case IRP_MN_QUERY_REMOVE_DEVICE:
		{ 
			returnStatus = STATUS_SUCCESS;
			break;
		}

		case IRP_MN_REMOVE_DEVICE :
		{ 
			IoDeleteDevice ( DeviceObject );
			returnStatus = STATUS_SUCCESS;
			break;
		}

		case IRP_MN_QUERY_BUS_INFORMATION:
		{
			Irp->IoStatus.Information = 0;
			returnStatus = STATUS_NOT_SUPPORTED;
			break;
		}

		case IRP_MN_QUERY_DEVICE_RELATIONS :
		{ 
			PDEVICE_RELATIONS pRelations;

			if( pStack->Parameters.QueryDeviceRelations.Type == TargetDeviceRelation )
			{
				pRelations = (PDEVICE_RELATIONS)ExAllocatePool( 
					PagedPool, 
					sizeof(ULONG) + sizeof(PDEVICE_OBJECT) * 2 );
				pRelations->Count = 1;
				pRelations->Objects[0] = DeviceObject;
				pRelations->Objects[1] = 0;

				ObReferenceObject( pRelations->Objects[0] );

				Irp->IoStatus.Information = (ULONG)pRelations;				
				returnStatus = STATUS_SUCCESS;
			}
			break;
		}

	case IRP_MN_QUERY_DEVICE_TEXT:
		{
			PWCHAR string=0;
			ULONG dwSize;
			PDEVICE_EXTENSION ParentDeviceExtension;

			ParentDeviceExtension = (PDEVICE_EXTENSION)deviceExtension->ParentDeviceObject->DeviceExtension;
			
			if( pStack->Parameters.QueryDeviceText.DeviceTextType != DeviceTextDescription  )
				break;

			if(( ParentDeviceExtension->ChildDeviceObject[0] == DeviceObject )
				|| ( ParentDeviceExtension->ChildDeviceObject[1] == DeviceObject ) )
			{
				string = (WCHAR *)ExAllocatePool( PagedPool, sizeof(CHILDDEVICETEXT)+2 );
				RtlZeroMemory( string, sizeof(CHILDDEVICETEXT)+2 );
				RtlMoveMemory( string, CHILDDEVICETEXT, sizeof(CHILDDEVICETEXT) );
				returnStatus = STATUS_SUCCESS;
			}
			else
				break;

			Irp->IoStatus.Information = (ULONG)string;				
			returnStatus = STATUS_SUCCESS;
			break;
		}

		case IRP_MN_QUERY_ID :
		{ 
			WCHAR * pID = NULL;
			PDEVICE_EXTENSION ParentDeviceExtension;

			ParentDeviceExtension = (PDEVICE_EXTENSION)deviceExtension->ParentDeviceObject->DeviceExtension;

			switch( pStack->Parameters.QueryId.IdType )
			{
				case BusQueryDeviceID:
					if(( ParentDeviceExtension->ChildDeviceObject[0] == DeviceObject )
						|| ( ParentDeviceExtension->ChildDeviceObject[1] == DeviceObject ))
					{
						pID = (WCHAR *)ExAllocatePool( PagedPool, sizeof(CHILDDEVICEDEVICEID)+2 );
						RtlZeroMemory( pID, sizeof(CHILDDEVICEDEVICEID)+2 );
						RtlMoveMemory( pID, CHILDDEVICEDEVICEID, sizeof(CHILDDEVICEDEVICEID) );
						returnStatus = STATUS_SUCCESS;
					}
					break;

				case BusQueryHardwareIDs:
					if(( ParentDeviceExtension->ChildDeviceObject[0] == DeviceObject )
						|| ( ParentDeviceExtension->ChildDeviceObject[1] == DeviceObject ))
					{
						pID = (WCHAR *)ExAllocatePool( PagedPool, sizeof(CHILDDEVICEHARDWAREID)+4 );
						RtlZeroMemory( pID, sizeof(CHILDDEVICEHARDWAREID)+4 );
						RtlMoveMemory( pID, CHILDDEVICEHARDWAREID, sizeof(CHILDDEVICEHARDWAREID) );
						returnStatus = STATUS_SUCCESS;
					}
					break;

				case BusQueryCompatibleIDs:
					if(( ParentDeviceExtension->ChildDeviceObject[0] == DeviceObject )
						|| ( ParentDeviceExtension->ChildDeviceObject[1] == DeviceObject ))
					{
						pID = (WCHAR *)ExAllocatePool( PagedPool, sizeof(CHILDDEVICECOMPATIBLEID)+4 );
						RtlZeroMemory( pID, sizeof(CHILDDEVICECOMPATIBLEID)+4 );
						RtlMoveMemory( pID, CHILDDEVICECOMPATIBLEID, sizeof(CHILDDEVICECOMPATIBLEID) );
						returnStatus = STATUS_SUCCESS;
					}
					break;
				case BusQueryInstanceID:
					if( ParentDeviceExtension->ChildDeviceObject[0] == DeviceObject )
					{
						pID = (WCHAR *)ExAllocatePool( PagedPool, sizeof(CHILDDEVICEINSTANCEID1)+2 );
						RtlZeroMemory( pID, sizeof(CHILDDEVICEINSTANCEID1)+2 );
						RtlMoveMemory( pID, CHILDDEVICEINSTANCEID1, sizeof(CHILDDEVICEINSTANCEID1) );
						returnStatus = STATUS_SUCCESS;
					}
					else if( ParentDeviceExtension->ChildDeviceObject[1] == DeviceObject )
					{
						pID = (WCHAR *)ExAllocatePool( PagedPool, sizeof(CHILDDEVICEINSTANCEID2)+2 );
						RtlZeroMemory( pID, sizeof(CHILDDEVICEINSTANCEID2)+2 );
						RtlMoveMemory( pID, CHILDDEVICEINSTANCEID2, sizeof(CHILDDEVICEINSTANCEID2) );
						returnStatus = STATUS_SUCCESS;
					}
					break;
			}
			Irp->IoStatus.Information = (ULONG)pID;				
			break;
		}
	} 

	Irp->IoStatus.Status = returnStatus;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );

	return returnStatus;
}

NTSTATUS
BUSDRIVER_PnpDispatch
	(
	IN PDEVICE_OBJECT DeviceObject,   
	IN PIRP Irp                       
	)
{
	PIO_STACK_LOCATION pStack; 
	PDEVICE_EXTENSION deviceExtension; 
	
	pStack = IoGetCurrentIrpStackLocation ( Irp );
	deviceExtension = DeviceObject->DeviceExtension;

	if( deviceExtension->DeviceObjectType == BUSDRIVERDEVICE )
	{ // BusDriverDevice
		return BusDriverDevice_PnpDispatch( DeviceObject, Irp );
	}
	else
	{ // ChildDriverDevice
		return ChildDriverDevice_PnpDispatch( DeviceObject, Irp );
	}			
}

VOID                                                                
BUSDRIVER_Unload                                                       
	(                                                               
	IN PDRIVER_OBJECT DriverObject 
	)
{
}

NTSTATUS                                                            
DriverEntry                                                         
	(                                                               
	IN PDRIVER_OBJECT DriverObject,             
	IN PUNICODE_STRING RegistryPath             
							
	)
{
	NTSTATUS returnStatus = STATUS_SUCCESS;    

	DriverObject->DriverUnload = BUSDRIVER_Unload;
	DriverObject->DriverExtension->AddDevice = BUSDRIVER_AddDevice;
	DriverObject->MajorFunction[IRP_MJ_PNP] = BUSDRIVER_PnpDispatch;

	return returnStatus;
}

