#include <ntddk.h>        /* required for WDM driver development */
#include <stdarg.h>     /* standard io include files */
#include <stdio.h>

#include "devioctl.h"
#include <initguid.h>

DEFINE_GUID(GUID_SAMPLE, 
0x5665dec0, 0xa40a, 0x11d1, 0xb9, 0x84, 0x0, 0x20, 0xaf, 0xd7, 0x97, 0x78);
//  츮  DeviceStack  Interface GUID մϴ

#include "SAMPLE1.h"    

void IrpStringOut( PIRP Irp ) // IRP Ʈմϴ
{
	KIRQL Irql;
	PIO_STACK_LOCATION pStack;
	
	pStack = IoGetCurrentIrpStackLocation( Irp );

	switch( pStack->MajorFunction )
	{
	case IRP_MJ_CREATE:
		DbgPrint("[IRP_MJ_CREATE]");
		break;
	case IRP_MJ_READ:
		DbgPrint("[IRP_MJ_READ]");
		break;
	case IRP_MJ_WRITE:
		DbgPrint("[IRP_MJ_WRITE]");
		break;
	case IRP_MJ_CLEANUP:
		DbgPrint("[IRP_MJ_CLEANUP]");
		break;
	case IRP_MJ_CLOSE:
		DbgPrint("[IRP_MJ_CLOSE]");
		break;
	case IRP_MJ_DEVICE_CONTROL:
		DbgPrint("[IRP_MJ_DEVICE_CONTROL]");
		DbgPrint("ControlCode = %8X", pStack->Parameters.DeviceIoControl.IoControlCode);
		break;
	case IRP_MJ_PNP:
		DbgPrint("[IRP_MJ_PNP]");
		switch( pStack->MinorFunction )
		{
			case IRP_MN_START_DEVICE:
				DbgPrint("IRP_MN_START_DEVICE");
				break;
			case IRP_MN_REMOVE_DEVICE:
				DbgPrint("IRP_MN_REMOVE_DEVICE");
				break;
			case IRP_MN_SURPRISE_REMOVAL:
				DbgPrint("IRP_MN_SURPRISE_REMOVAL");
				break;
			default:
				DbgPrint("ANY MN FUNCTION");
				break;
		}
		break;
	case IRP_MJ_POWER:
		DbgPrint("[IRP_MJ_POWER]");
		switch( pStack->MinorFunction )
		{
			case IRP_MN_SET_POWER:
				DbgPrint("IRP_MN_SET_POWER");
				break;
			case IRP_MN_QUERY_POWER:
				DbgPrint("IRP_MN_QUERY_POWER");
				break;
			default:
				DbgPrint("ANY MN FUNCTION");
				break;
		}
		break;
	default:
		DbgPrint("[ANY IRP] = 0x%8X\n", Irp );
		return;
	}
	DbgPrint("\n");
}

NTSTATUS                                                            
DriverEntry                                                         
	(                                                               
	IN PDRIVER_OBJECT DriverObject,             /* pointer to the device object */
	IN PUNICODE_STRING RegistryPath             /* pointer to a unicode string representing the path */
												/*    to the drivers specific key in the registry    */
	)// ̹ ޸𸮿 óϴ 쿡 ѹ ȣ˴ϴ
{
	NTSTATUS returnStatus = STATUS_SUCCESS;     /* stores status to be returned */

	DbgPrint("DriverEntry ++ \n" );

	DriverObject->DriverUnload = SAMPLE1_Unload;
	DriverObject->DriverExtension->AddDevice = SAMPLE1_AddDevice;
	DriverObject->MajorFunction[IRP_MJ_PNP] = SAMPLE1_PnpDispatch;
	DriverObject->MajorFunction[IRP_MJ_POWER] = SAMPLE1_PowerDispatch;
	DriverObject->MajorFunction[IRP_MJ_CREATE] = SAMPLE1_Create; // Win32 API CreateFile() ˴ϴ
	DriverObject->MajorFunction[IRP_MJ_CLOSE] = SAMPLE1_Close; // Win32 API CloseHandle() ˴ϴ
	DriverObject->MajorFunction[IRP_MJ_CLEANUP] = SAMPLE1_Cleanup; // Win32 API CloseHandle() ˴ϴ

	DbgPrint("DriverEntry -- \n" );
	return returnStatus;
}

NTSTATUS
SAMPLE1_AddDevice
	(
	IN PDRIVER_OBJECT DriverObject,         
	IN PDEVICE_OBJECT PhysicalDeviceObject  
	)// PNPManager ̰ ؼ IRP DeviceStack ϶ û մϴ
{
	/* IRP DeviceStack ϱ ؼ  ܰ踦 ġ ˴ϴ 
	1. DeviceObject۾
	2. ϴ DeviceStack DeviceObject ø
	3. DeviceObject Flag ϱ
	*/
	PDEVICE_EXTENSION deviceExtension;              
	NTSTATUS returnStatus = STATUS_SUCCESS;         
	PDEVICE_OBJECT DeviceObject = NULL;             
	ULONG dwRet;

	DbgPrint("SAMPLE1_AddDevice ++ \n" );

	returnStatus = IoCreateDevice                                       
					(                                               
						DriverObject,                               
						sizeof ( DEVICE_EXTENSION ),                
						NULL,                   
						FILE_DEVICE_UNKNOWN,
						FILE_AUTOGENERATED_DEVICE_NAME,                                          
						FALSE,                                      
						&DeviceObject       
					);// DeviceObject մϴ

	if( !NT_SUCCESS( returnStatus ) )
	{
		DbgPrint("IoCreateDevice() Error, Status = %8X\n", returnStatus );
		goto SAMPLE1_AddDevice_Exit;
	}

	deviceExtension = DeviceObject->DeviceExtension;

	RtlZeroMemory( deviceExtension, sizeof(DEVICE_EXTENSION));

	deviceExtension->PhysicalDeviceObject = PhysicalDeviceObject;
	
	deviceExtension->DeviceObject = DeviceObject;

	deviceExtension->NextLayerDeviceObject =
	IoAttachDeviceToDeviceStack (
							DeviceObject,
							PhysicalDeviceObject	
	);// ϴ DeviceStack DeviceObject øϴ

	DeviceObject->Flags |= deviceExtension->NextLayerDeviceObject->Flags & 
							( DO_POWER_PAGABLE  | DO_POWER_INRUSH); // DeviceObject Flag մϴ

	DeviceObject->Flags &= ~DO_DEVICE_INITIALIZING; // DeviceObject Flag մϴ

	IoRegisterDeviceInterface( PhysicalDeviceObject, &GUID_SAMPLE, NULL, &deviceExtension->UnicodeString );
	//  츮  DeviceStack  InterfaceGuid մϴ
	// ϵǴ de->UnicodeString ڰ ϴ ̸Դϴ


SAMPLE1_AddDevice_Exit:
	DbgPrint("SAMPLE1_AddDevice -- \n" );
	return returnStatus;
}

VOID                                                                
SAMPLE1_Unload                                                       
	(                                                               
	IN PDRIVER_OBJECT DriverObject  /* pointer to device object */
	)
{// ̻ ̹  DeviceObject  ̰ ȣ, ̹ ޸𸮿 ŵ˴ϴ

	DbgPrint("SAMPLE1_Unload ++ \n" );

	// Do job..

	DbgPrint("SAMPLE1_Unload -- \n" );
}

NTSTATUS
SAMPLE1_DeferIrpCompletion
	(
	IN PDEVICE_OBJECT DeviceObject,     /* pointer to a device object */
	IN PIRP Irp,                        /* pointer to a I/O request packet */
	IN PVOID Context                    /* driver defined context */
	)// ɾ IRP NextLayer ̹ IoCompleteRequest()Լ ؼ ûǴ 쿡 ȣ˴ϴ
{
	PKEVENT event = Context;    /* kernel event passed in Context */

	DbgPrint("SAMPLE1_DeferIrpCompletion ++ \n" );

	KeSetEvent(event,
			   1,
			   FALSE);

	
	DbgPrint("SAMPLE1_DeferIrpCompletion -- \n" );

	return STATUS_MORE_PROCESSING_REQUIRED; // IoCompleteRequest()Լ Ͽ,  IRP  ߴ϶ û մϴ
}

NTSTATUS
SAMPLE1_PnpDispatch
	(
	IN PDEVICE_OBJECT DeviceObject,     /* pointer to device object */
	IN PIRP Irp                         /* pointer to an I/O request packet */
	)
{// PNPManager  IRP óϴ ó
	PIO_STACK_LOCATION pStack;                /* pointer to IRP stack */
	PDEVICE_EXTENSION deviceExtension;          /* pointer to device extention */
	NTSTATUS returnStatus = STATUS_SUCCESS;     /* stores status to be returned */
	PDEVICE_OBJECT NextLayerDeviceObject;           /* pointer to the stack device object */
	
	DbgPrint("SAMPLE1_PnpDispatch ++ \n" );

	IrpStringOut( Irp );

	pStack = IoGetCurrentIrpStackLocation ( Irp );
	deviceExtension = DeviceObject->DeviceExtension;
	NextLayerDeviceObject = deviceExtension->NextLayerDeviceObject;
			
	switch ( pStack->MinorFunction )
	{
		case IRP_MN_START_DEVICE :
		{// DeviceStack ϶ 㰡 ǹմϴ
			KEVENT event;   /* event used to wait for an operation to complete */
			KeInitializeEvent(&event, NotificationEvent, FALSE);
			IoCopyCurrentIrpStackLocationToNext(Irp);  
			IoSetCompletionRoutine(Irp,
								   SAMPLE1_DeferIrpCompletion,
								   &event,
								   TRUE,
								   TRUE,
								   TRUE); // SAMPLE1_DeferIrpCompletion()Լ  IRP  Next Layer Driver ȣϴ
										  // IoCompleteRequest()Լ ؼ ȣǵ մϴ
			
			returnStatus = IoCallDriver(NextLayerDeviceObject,Irp); // NextLayer Driver   ѵ,  Ȯմϴ
			if (returnStatus == STATUS_PENDING) 
			{
				KeWaitForSingleObject(
					&event,
					Executive,
					KernelMode,
					FALSE,
					NULL);
				returnStatus = Irp->IoStatus.Status;
			}
			if( !NT_SUCCESS( returnStatus ) )
			{
				DbgPrint("Start Device Fail\n" );
				IoCompleteRequest( Irp, IO_NO_INCREMENT );
				goto SAMPLE1_PnpDispatch_Exit;
			}
			// NextLayer Driver    ٷ, 츮 ΰ ص ٴ ǹԴϴ

			// Do Job...
			IoSetDeviceInterfaceState( &deviceExtension->UnicodeString, TRUE ); 
			// ڰ 츮 DeviceStack ϴ  

			IoCompleteRequest( Irp, IO_NO_INCREMENT ); //  IRP ûմϴ
			// SAMPLE1_DeferIrpCompletion()Լ 츮 IRP  Next Layer Driver ȣ IoCompleteRequest() 
		}	// ϵ "STATUS_MORE_PROCESSING_REQUIRED" ױ⶧, 츮 û ؾմϴ
		break;

		case IRP_MN_SURPRISE_REMOVAL :
		{ // ڱ ġ PC ŵǾٴ ǹ̿ ޵˴ϴ

			IoSetDeviceInterfaceState( &deviceExtension->UnicodeString, FALSE );
			// ڰ 츮 DeviceStack ϴ  

			Irp->IoStatus.Status = STATUS_SUCCESS;
			IoSkipCurrentIrpStackLocation( Irp ); //  IRP  ɶ ̻   ʰڴٴ ǹԴϴ
			returnStatus = IoCallDriver(NextLayerDeviceObject, Irp);
		}
		break;

		case IRP_MN_REMOVE_DEVICE :
		{ // IRP DeviceStack ü϶ û ǹ̷ ޵˴ϴ
			IoSetDeviceInterfaceState( &deviceExtension->UnicodeString, FALSE );
			// ڰ 츮 DeviceStack ϴ  

			RtlFreeUnicodeString( &deviceExtension->UnicodeString );
			// ڰ ϴ ̸ ޸𸮿 Ѵ

			IoDetachDevice ( NextLayerDeviceObject );// Next Layer Driver  켱 ϴ
													 // SAMPLE1_AddDevice() IoAttachDeviceToDeviceStack()Լ ݴǹ
			IoDeleteDevice ( DeviceObject );		 // 츮 DeviceObject մϴ
													 // SAMPLE1_AddDevice() IoCreateDevice()Լ ݴǹ
			Irp->IoStatus.Status = STATUS_SUCCESS;
			IoSkipCurrentIrpStackLocation( Irp ); //  IRP  ɶ ̻   ʰڴٴ ǹԴϴ
			returnStatus = IoCallDriver(NextLayerDeviceObject, Irp);
		}
		break;

		default:
		{
			IoSkipCurrentIrpStackLocation( Irp ); //  IRP  ɶ ̻   ʰڴٴ ǹԴϴ
			returnStatus = IoCallDriver(NextLayerDeviceObject, Irp);
		}
		break;
	} 

SAMPLE1_PnpDispatch_Exit:
	DbgPrint("SAMPLE1_PnpDispatch -- \n" );

	return returnStatus;
}

NTSTATUS                                                            
SAMPLE1_PowerDispatch
	(                                                               
		IN PDEVICE_OBJECT DeviceObject,                             
		IN PIRP Irp                                                 
	)
{// POWERManager  IRP óϴ ó
	PDEVICE_EXTENSION		deviceExtension;
	NTSTATUS returnStatus;
	PDEVICE_OBJECT NextLayerDeviceObject;           /* pointer to the stack device object */
	
	DbgPrint("SAMPLE1_PowerDispatch ++ \n" );

	IrpStringOut( Irp );

	deviceExtension = DeviceObject->DeviceExtension;
	NextLayerDeviceObject = deviceExtension->NextLayerDeviceObject;

	PoStartNextPowerIrp( Irp ); //  ٸ Power IRP   ִٰ PowerManager ˸ϴ
	IoSkipCurrentIrpStackLocation( Irp ); //  IRP  ɶ ̻   ʰڴٴ ǹԴϴ
	returnStatus = PoCallDriver(NextLayerDeviceObject, Irp); // Power IRP ٸ IRP ޸, ݵ PoCallDriver()Լ ؾ մϴ
	
	DbgPrint("SAMPLE1_PowerDispatch -- \n" );
	return returnStatus;
} 

NTSTATUS                                                            
SAMPLE1_Create
	(                                                               
		IN PDEVICE_OBJECT DeviceObject,                             
		IN PIRP Irp                                                 
	)
{// Application  CreateFile()Լ Ǵ ó
	NTSTATUS returnStatus = STATUS_SUCCESS;

	DbgPrint("SAMPLE1_Create ++ \n" );

	IrpStringOut( Irp );
	
	DbgPrint("SAMPLE1_Create -- \n" );

	Irp->IoStatus.Status = returnStatus;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );

	return returnStatus;
} 

NTSTATUS                                                            
SAMPLE1_Cleanup
	(                                                               
		IN PDEVICE_OBJECT DeviceObject,                             
		IN PIRP Irp                                                 
	)
{// Application  CloseHandle()Լ Ǵ ó
	NTSTATUS returnStatus = STATUS_SUCCESS;

	DbgPrint("SAMPLE1_Cleanup ++ \n" );

	IrpStringOut( Irp );
	
	DbgPrint("SAMPLE1_Cleanup -- \n" );

	Irp->IoStatus.Status = returnStatus;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );

	return returnStatus;
} 

NTSTATUS                                                            
SAMPLE1_Close
	(                                                               
		IN PDEVICE_OBJECT DeviceObject,                             
		IN PIRP Irp                                                 
	)
{// Application  CloseHandle()Լ Ǵ ó
	NTSTATUS returnStatus = STATUS_SUCCESS;

	DbgPrint("SAMPLE1_Close ++ \n" );

	IrpStringOut( Irp );
	
	DbgPrint("SAMPLE1_Close -- \n" );

	Irp->IoStatus.Status = returnStatus;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );

	return returnStatus;
} 
