#include <ntifs.H>
#include <F_IFACE.H>
#include "filehiddenroutine.h"

ULONG g_HiddenFileAction;

WCHAR * g_HiddenFileList[MAXHIDDENFILEENTRIES+1];

NTSTATUS InitializeHiddenFileList()
{
	int i;
	for( i = 0 ; i < MAXHIDDENFILEENTRIES ; i ++ )
	{
		g_HiddenFileList[ i ] = (WCHAR *)ExAllocatePool( NonPagedPool, MAXHIDDENFILENAMESIZE * sizeof(WCHAR) );
		if( !g_HiddenFileList[ i ] )
			return STATUS_UNSUCCESSFUL;
		RtlZeroMemory( g_HiddenFileList[ i ], MAXHIDDENFILENAMESIZE * sizeof(WCHAR) );
	}
	return STATUS_SUCCESS;
}

ULONG IsExistHiddenFileList( WCHAR *pwFilename )
{
	int i;
	for( i = 0 ; i < MAXHIDDENFILEENTRIES ; i ++ )
	{
		if( !g_HiddenFileList[ i ] )
			continue;
		
		if( _wcsicmp( g_HiddenFileList[ i ], pwFilename ) )
			continue;

		return TRUE;
	}
	return FALSE;
}

ULONG IsExistCurrentEntryForHiddenFile( PFILE_BOTH_DIR_INFORMATION pInfo, PWCHAR pwPrefixName )
{
	WCHAR * pdwTempPath;
	ULONG bRet;

	pdwTempPath = (WCHAR *)ExAllocatePool( NonPagedPool, MAXHIDDENFILENAMESIZE * sizeof(WCHAR) );
	RtlZeroMemory( pdwTempPath, MAXHIDDENFILENAMESIZE * sizeof(WCHAR));
	
	wcscpy( pdwTempPath, pwPrefixName );
	if( pdwTempPath[ wcslen(pdwTempPath) - 1] != L'\\' )
		wcscat( pdwTempPath, L"\\" );

	memcpy( pdwTempPath + wcslen(pdwTempPath), pInfo->FileName, pInfo->FileNameLength );

	bRet = IsExistHiddenFileList(pdwTempPath);

	ExFreePool( pdwTempPath );
	return bRet;
}

ULONG HideFiles( PWCHAR pwPrefixName, PFILE_BOTH_DIR_INFORMATION pInfo, ULONG dwInfoSize )
{
	PFILE_BOTH_DIR_INFORMATION pBackupInfo, pBackupInfo_for_Delete, pBackupInfo_for_LastEntry;
	PFILE_BOTH_DIR_INFORMATION pBackup_pInfo;
	
	ULONG bRet;
	ULONG dwCopySize=0;
	ULONG dwOriginalInfoLastEntrySize;

	if( g_HiddenFileAction == FALSE )
		return dwInfoSize;
	
	dwOriginalInfoLastEntrySize = dwInfoSize;
	pBackup_pInfo = pInfo;
		
	pBackupInfo = (PFILE_BOTH_DIR_INFORMATION)ExAllocatePool( NonPagedPool, dwInfoSize );
	if( !pBackupInfo )
	{
		return STATUS_SUCCESS;;
	}
		
	pBackupInfo_for_Delete = pBackupInfo;
	pBackupInfo_for_LastEntry = (PFILE_BOTH_DIR_INFORMATION)0;
	
	RtlZeroMemory( pBackupInfo, dwInfoSize );

	while( pInfo->NextEntryOffset )
	{
		// Ʈ ƴϸ..
		dwOriginalInfoLastEntrySize -= pInfo->NextEntryOffset;
		
		bRet = IsExistCurrentEntryForHiddenFile( pInfo, pwPrefixName );
		if( bRet == FALSE )
		{
			// ص 
			memcpy( pBackupInfo, pInfo, pInfo->NextEntryOffset );
			dwCopySize += pInfo->NextEntryOffset;
			pBackupInfo_for_LastEntry = pBackupInfo;
			pBackupInfo = (PFILE_BOTH_DIR_INFORMATION)((PCHAR)pBackupInfo + (ULONG)(pBackupInfo->NextEntryOffset));
			pInfo = (PFILE_BOTH_DIR_INFORMATION)((PCHAR)pInfo + (ULONG)(pInfo->NextEntryOffset));
			continue;
		}

		// ߰ߵǸ  Ʈ ϸ ȵȴ
		pInfo = (PFILE_BOTH_DIR_INFORMATION)((PCHAR)pInfo + (ULONG)(pInfo->NextEntryOffset));
		continue;
	}

	// Ʈó
	bRet = IsExistCurrentEntryForHiddenFile( pInfo, pwPrefixName );
	if( bRet == FALSE )
	{
		// ص 
		memcpy( pBackupInfo, pInfo,  dwOriginalInfoLastEntrySize );
		dwCopySize += dwOriginalInfoLastEntrySize;
	}
	else
	{
		// ߰ߵǸ  Ʈ ϸ ȵȴ
		// Ʈ ó
		if( pBackupInfo_for_LastEntry )
			pBackupInfo_for_LastEntry->NextEntryOffset = 0;
	}	

	pBackupInfo = pBackupInfo_for_Delete;
		
	if( dwInfoSize != dwCopySize )
	{
		RtlZeroMemory( pBackup_pInfo, dwInfoSize );
		memcpy( pBackup_pInfo, pBackupInfo, dwCopySize );
	}

	ExFreePool( pBackupInfo );
	return dwCopySize;
}	

NTSTATUS HiddenFilesRoutine( PDEVICE_OBJECT pDevObj, PIRP pIrp )
{
	PIO_STACK_LOCATION pCurIrpStack;
	NTSTATUS ntStatus;
	ULONG Information;
	ULONG ulIoControlCode;
	PVOID pInputBuffer, pOutputBuffer;
	ULONG ulInputBufferLength, ulOutputBufferLength;
	
	pIrp->IoStatus.Information = 0;
	
	pCurIrpStack = IoGetCurrentIrpStackLocation(pIrp);

    	ulIoControlCode = pCurIrpStack->Parameters.DeviceIoControl.IoControlCode;
    
    	pInputBuffer = pIrp->AssociatedIrp.SystemBuffer;
	pOutputBuffer = pIrp->AssociatedIrp.SystemBuffer;
    	ulInputBufferLength = pCurIrpStack->Parameters.DeviceIoControl.InputBufferLength;
    	ulOutputBufferLength = pCurIrpStack->Parameters.DeviceIoControl.OutputBufferLength;
	
	ntStatus = STATUS_UNSUCCESSFUL;
	Information = 0;
	
	switch(ulIoControlCode) 
	{
		case IOCTL_HIDDEN_FILES:
			g_HiddenFileAction = TRUE;
			ntStatus = STATUS_SUCCESS;
			break;
		case IOCTL_SHOW_FILES:
			g_HiddenFileAction = FALSE;
			ntStatus = STATUS_SUCCESS;
			break;
		case IOCTL_ADD_HIDDEN_FILE:
		{
			int i;
			PADD_HIDDEN_FILE pEntry;
			PHIDDEN_FILE_HANDLE pReturnHandle;
			
			pEntry = (PADD_HIDDEN_FILE)pInputBuffer;
			pReturnHandle = (PHIDDEN_FILE_HANDLE)pOutputBuffer;

			for( i = 0 ; i < MAXHIDDENFILEENTRIES ; i ++ )
			{
				if( g_HiddenFileList[ i ][0] )
					continue;
				memcpy( g_HiddenFileList[ i ], &pEntry->Filename[0], MAXHIDDENFILENAMESIZE * sizeof(WCHAR) );
				break;				
			}
			if( i >= MAXHIDDENFILEENTRIES )
			{
				break;
			}
			pReturnHandle->Handle = i;
			ntStatus = STATUS_SUCCESS;
			Information = sizeof(HIDDEN_FILE_HANDLE);
		}
			break;
		case IOCTL_DEL_HIDDEN_FILE:
		{
			int i;
			PDEL_HIDDEN_FILE pEntry;
			
			pEntry = (PDEL_HIDDEN_FILE)pInputBuffer;
			if( (int)pEntry->Handle.Handle >= MAXHIDDENFILEENTRIES )
				break;

			RtlZeroMemory( g_HiddenFileList[ pEntry->Handle.Handle ], MAXHIDDENFILENAMESIZE * sizeof(WCHAR) );
			ntStatus = STATUS_SUCCESS;
		}
			break;
		case IOCTL_ALL_RESET_HIDDEN_FILES:
		{
			int i;
			for( i = 0 ; i < MAXHIDDENFILEENTRIES ; i ++ )
			{
				RtlZeroMemory( g_HiddenFileList[ i ], MAXHIDDENFILENAMESIZE * sizeof(WCHAR) );
			}
		}
				
			break;
	}
	pIrp->IoStatus.Status = ntStatus;
	pIrp->IoStatus.Information = Information;
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );
	return ntStatus;
}

ULONG GetRealSize( PFILE_BOTH_DIR_INFORMATION pInfo )
{
	ULONG dwCopySize=0;

	while( pInfo->NextEntryOffset )
	{
		// Ʈ ƴϸ..
		dwCopySize += pInfo->NextEntryOffset;
		pInfo = (PFILE_BOTH_DIR_INFORMATION)((PCHAR)pInfo + (ULONG)(pInfo->NextEntryOffset));
		continue;
	}

	dwCopySize += (sizeof( FILE_BOTH_DIR_INFORMATION ) - 2 + pInfo->FileNameLength);
	return dwCopySize;
}	
	
