#include <ntddk.h>

#include "util.h"

void SetRegValue( PWCHAR pEntryKey, PUNICODE_STRING pRegKey, PWCHAR pKeyValue, ULONG dwSize )
{
	PWCHAR pRegKeyToRead;
	pRegKeyToRead = ExAllocatePool(PagedPool, 255);
	RtlZeroMemory( pRegKeyToRead, 255 );
	RtlMoveMemory( pRegKeyToRead, pRegKey->Buffer, pRegKey->Length );
	RtlWriteRegistryValue( RTL_REGISTRY_ABSOLUTE|RTL_REGISTRY_OPTIONAL, pRegKeyToRead, pEntryKey, REG_SZ, pKeyValue, dwSize );
	ExFreePool( pRegKeyToRead );
}

NTSTATUS RegisterSerialClassDriver( PDEVICE_OBJECT DeviceObject )
{
	NTSTATUS ntStatus;
	HANDLE pnpKey;
	unsigned char TempBuffer[255];
	KEY_VALUE_FULL_INFORMATION * pBuffer;
	UNICODE_STRING PortName, linkName, devName, uniRegPath, uniRegValue;
	WCHAR DOS_NAME[255]=L"\\DosDevices\\";
	WCHAR cPortName[255];
	ULONG dwRet;
	ULONG i;
	WCHAR PropertyBuffer[255];
	PDEVICE_EXTENSION deviceExtension;

	deviceExtension = DeviceObject->DeviceExtension;

	ntStatus = IoOpenDeviceRegistryKey( deviceExtension->PhysicalDeviceObject, PLUGPLAY_REGKEY_DEVICE, STANDARD_RIGHTS_WRITE, &pnpKey );
	if(!NT_SUCCESS(ntStatus))
		return ntStatus;
	pBuffer = (KEY_VALUE_FULL_INFORMATION *)TempBuffer;
	pBuffer->NameLength = 255;
	RtlInitUnicodeString( &PortName, L"PortName" );
	ntStatus = ZwQueryValueKey(pnpKey, &PortName, KeyValueFullInformation, pBuffer, 255, &dwRet );
	if(NT_SUCCESS(ntStatus))
	{
		RtlZeroMemory( DOS_NAME, 255 * sizeof(WCHAR) );
		RtlZeroMemory( deviceExtension->DOS_NAME, 255 * sizeof(WCHAR) );
		wcscat( DOS_NAME, L"\\DosDevices\\" );
		memcpy( (PUCHAR)(DOS_NAME + wcslen( DOS_NAME )), ((PUCHAR)TempBuffer)+pBuffer->DataOffset, pBuffer->DataLength );
		wcscpy( deviceExtension->DOS_NAME, DOS_NAME );
		wcscat(DOS_NAME, (unsigned short *)L":" );
		wcscpy( deviceExtension->DOS_NAME2, DOS_NAME );
		
		RtlCopyMemory( ((PUCHAR)cPortName), ((PUCHAR)TempBuffer)+pBuffer->DataOffset, pBuffer->DataLength );
		ZwClose(pnpKey);

		ntStatus = IoGetDeviceProperty( deviceExtension->PhysicalDeviceObject, DevicePropertyPhysicalDeviceObjectName 
			, 255, PropertyBuffer, &dwRet );
		if(!NT_SUCCESS(ntStatus))
			return ntStatus;

		RtlInitUnicodeString( &devName, PropertyBuffer );
		RtlInitUnicodeString( &linkName, deviceExtension->DOS_NAME );
		IoCreateSymbolicLink( &linkName, &devName );
		RtlInitUnicodeString( &linkName, deviceExtension->DOS_NAME2 );
		IoCreateSymbolicLink( &linkName, &devName );
		
		RtlInitUnicodeString( &uniRegPath, DRIVER_REGPATH );
		RtlInitUnicodeString( &uniRegValue, cPortName );
		SetRegValue( PropertyBuffer, &uniRegPath, cPortName, uniRegValue.Length );
		IoSetDeviceInterfaceState( &deviceExtension->UnicodeString, TRUE );
	}
	else
	{
		ZwClose(pnpKey);
	}
	return ntStatus;
}

PIRP OutputPendingWaitIrp( PDEVICE_EXTENSION deviceExtension )
{
	PIRP Irp;
	KIRQL OldIrql;

	KeRaiseIrql( DISPATCH_LEVEL, &OldIrql );
	
	if( deviceExtension->WaitQHead == deviceExtension->WaitQTail )
	{
		KeLowerIrql( OldIrql );
		return 0;
	}
	Irp = deviceExtension->PendingWaitIrp[deviceExtension->WaitQHead];
	deviceExtension->WaitQHead = ((deviceExtension->WaitQHead+1) >= MAXPENDINGIRP )? 0:deviceExtension->WaitQHead + 1;
	KeLowerIrql( OldIrql );
	return Irp;
}

void InputPendingWaitIrp( PDEVICE_EXTENSION deviceExtension, PIRP Irp )
{
	ULONG dwTail;
	KIRQL OldIrql;

	KeRaiseIrql( DISPATCH_LEVEL, &OldIrql );
		
	dwTail = deviceExtension->WaitQTail+1;
	if( dwTail >= MAXPENDINGIRP )
		dwTail = 0;
	if( deviceExtension->WaitQHead == dwTail )
	{
		KeLowerIrql( OldIrql );
		return;
	}
	deviceExtension->PendingWaitIrp[deviceExtension->WaitQTail] = Irp;
	deviceExtension->WaitQTail = dwTail;
	KeLowerIrql( OldIrql );
	return;
}

ULONG SizeofPendingWaitIrp(PDEVICE_EXTENSION deviceExtension )
{
	KIRQL OldIrql;

	KeRaiseIrql( DISPATCH_LEVEL, &OldIrql );
		
	if( deviceExtension->WaitQHead == deviceExtension->WaitQTail )
	{
		KeLowerIrql( OldIrql );
		return 0;
	}
	if( deviceExtension->WaitQHead < deviceExtension->WaitQTail )
	{
		KeLowerIrql( OldIrql );
		return deviceExtension->WaitQTail - deviceExtension->WaitQHead;
	}
	else
	{
		KeLowerIrql( OldIrql );
		return (MAXPENDINGIRP - deviceExtension->WaitQHead + deviceExtension->WaitQTail);
	}
}

