#include <ntddk.h>        /* required for WDM driver development */
#include <stdarg.h>     /* standard io include files */
#include <stdio.h>

#include "devioctl.h"
#include <initguid.h>

DEFINE_GUID(GUID_SAMPLE, 
0x5665dec0, 0xa40a, 0x11d1, 0xb9, 0x84, 0x0, 0x20, 0xaf, 0xd7, 0x97, 0x78);
//  츮  DeviceStack  Interface GUID մϴ

// α׷ ְ ControlCode մϴ
#define IOCTL_SAMPLE_SET_FREQUENCY	CTL_CODE(FILE_DEVICE_UNKNOWN, 0x0800, METHOD_BUFFERED, FILE_ANY_ACCESS)

#pragma optimize("", off)
#include "ISR_DPC.h"    

#pragma pack(1)
typedef struct _CONFIG_ADDRESS
{
	union{
		struct{
			UCHAR Reserved1:2;
			UCHAR RegisterNumber:6;
			UCHAR FunctionNumber:3;
			UCHAR DeviceNumber:5;
			unsigned char BusNumber;
			unsigned char Reserved2;
		}bits;
		ULONG AsULONG;
	}u;
}CONFIG_ADDRESS, *PCONFIG_ADDRESS;
#pragma pack()

NTHALAPI ULONG
HalGetInterruptVector(
    IN INTERFACE_TYPE  InterfaceType,
    IN ULONG BusNumber,
    IN ULONG BusInterruptLevel,
    IN ULONG BusInterruptVector,
    OUT PKIRQL Irql,
    OUT PKAFFINITY Affinity
    );

void 
_SpeakerOn( unsigned short Frequency )
{
	unsigned short TimerValue;
	unsigned char TimerValueHighByte;
	unsigned char TimerValueLowByte;

	TimerValue = 1193180 / Frequency;
	TimerValueHighByte = TimerValue / 0x100;
	TimerValueLowByte = TimerValue % 0x100;

	WRITE_PORT_UCHAR( (PUCHAR)0x61, 0x00 );
	WRITE_PORT_UCHAR( (PUCHAR)0x43, 0xB6 );
	WRITE_PORT_UCHAR( (PUCHAR)0x42, TimerValueLowByte );
	WRITE_PORT_UCHAR( (PUCHAR)0x42, TimerValueHighByte );
	WRITE_PORT_UCHAR( (PUCHAR)0x61, 0x03 );
}

void 
_SpeakerOff()
{
	WRITE_PORT_UCHAR( (PUCHAR)0x61, 0x00 );
}

NTSTATUS                                                            
DriverEntry                                                         
	(                                                               
	IN PDRIVER_OBJECT DriverObject,             /* pointer to the device object */
	IN PUNICODE_STRING RegistryPath             /* pointer to a unicode string representing the path */
												/*    to the drivers specific key in the registry    */
	)// ̹ ޸𸮿 óϴ 쿡 ѹ ȣ˴ϴ
{
	NTSTATUS returnStatus = STATUS_SUCCESS;     /* stores status to be returned */

	DbgPrint("DriverEntry ++ \n" );

	DriverObject->DriverUnload = ISR_DPC_Unload;
	DriverObject->DriverExtension->AddDevice = ISR_DPC_AddDevice;
	DriverObject->MajorFunction[IRP_MJ_PNP] = ISR_DPC_PnpDispatch;
	DriverObject->MajorFunction[IRP_MJ_POWER] = ISR_DPC_PowerDispatch;
	DriverObject->MajorFunction[IRP_MJ_CREATE] = ISR_DPC_Create;
	DriverObject->MajorFunction[IRP_MJ_CLOSE] = ISR_DPC_Close;
	DriverObject->MajorFunction[IRP_MJ_CLEANUP] = ISR_DPC_Cleanup;
	// Win32 API CloseHandle();
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = ISR_DPC_DeviceIoControl;
	// Win32 API DeviceIoControl();
	DriverObject->DriverStartIo = ISR_DPC_StartIo; // <<<--- StartIo()Լ ϰ ֽϴ

	DbgPrint("DriverEntry -- \n" );
	return returnStatus;
}

NTSTATUS
ISR_DPC_AddDevice
	(
	IN PDRIVER_OBJECT DriverObject,         
	IN PDEVICE_OBJECT PhysicalDeviceObject  
	)// PNPManager ̰ ؼ IRP DeviceStack ϶ û մϴ
{
	/* IRP DeviceStack ϱ ؼ  ܰ踦 ġ ˴ϴ 
	1. DeviceObject۾
	2. ϴ DeviceStack DeviceObject ø
	3. DeviceObject Flag ϱ
	*/
	PDEVICE_EXTENSION deviceExtension;              
	NTSTATUS returnStatus = STATUS_SUCCESS;         
	PDEVICE_OBJECT DeviceObject = NULL;             
	ULONG dwRet;

	DbgPrint("ISR_DPC_AddDevice ++ \n" );

	returnStatus = IoCreateDevice                                       
					(                                               
						DriverObject,                               
						sizeof ( DEVICE_EXTENSION ),                
						NULL,                   
						FILE_DEVICE_UNKNOWN,
						FILE_AUTOGENERATED_DEVICE_NAME,                                          
						FALSE,                                      
						&DeviceObject       
					);// DeviceObject մϴ

	if( !NT_SUCCESS( returnStatus ) )
	{
		DbgPrint("IoCreateDevice() Error, Status = %8X\n", returnStatus );
		goto ISR_DPC_AddDevice_Exit;
	}

	deviceExtension = DeviceObject->DeviceExtension;

	RtlZeroMemory( deviceExtension, sizeof(DEVICE_EXTENSION));

	deviceExtension->PhysicalDeviceObject = PhysicalDeviceObject;
	
	deviceExtension->DeviceObject = DeviceObject;

	deviceExtension->NextLayerDeviceObject =
	IoAttachDeviceToDeviceStack (
							DeviceObject,
							PhysicalDeviceObject	
	);// ϴ DeviceStack DeviceObject øϴ

	DeviceObject->Flags |= deviceExtension->NextLayerDeviceObject->Flags & 
							( DO_POWER_PAGABLE  | DO_POWER_INRUSH); // DeviceObject Flag մϴ

	DeviceObject->Flags &= ~DO_DEVICE_INITIALIZING; // DeviceObject Flag մϴ

	// Timer DPC غմϴ
	KeInitializeTimer( &deviceExtension->MyTimer );
	KeInitializeDpc( &deviceExtension->MyDpc, MyDpcRoutine, DeviceObject );
	
	IoRegisterDeviceInterface( PhysicalDeviceObject, &GUID_SAMPLE, NULL, &deviceExtension->UnicodeString );
	//  츮  DeviceStack  InterfaceGuid մϴ
	// ϵǴ de->UnicodeString ڰ ϴ ̸Դϴ


ISR_DPC_AddDevice_Exit:
	DbgPrint("ISR_DPC_AddDevice -- \n" );
	return returnStatus;
}

VOID                                                                
ISR_DPC_Unload                                                       
	(                                                               
	IN PDRIVER_OBJECT DriverObject  /* pointer to device object */
	)
{// ̻ ̹  DeviceObject  ̰ ȣ, ̹ ޸𸮿 ŵ˴ϴ

	DbgPrint("ISR_DPC_Unload ++ \n" );

	// Do job..

	DbgPrint("ISR_DPC_Unload -- \n" );
}

NTSTATUS
ISR_DPC_DeferIrpCompletion
	(
	IN PDEVICE_OBJECT DeviceObject,     /* pointer to a device object */
	IN PIRP Irp,                        /* pointer to a I/O request packet */
	IN PVOID Context                    /* driver defined context */
	)// ɾ IRP NextLayer ̹ IoCompleteRequest()Լ ؼ ûǴ 쿡 ȣ˴ϴ
{
	PKEVENT event = Context;    /* kernel event passed in Context */

	DbgPrint("ISR_DPC_DeferIrpCompletion ++ \n" );

	KeSetEvent(event,
			   1,
			   FALSE);

	
	DbgPrint("ISR_DPC_DeferIrpCompletion -- \n" );

	return STATUS_MORE_PROCESSING_REQUIRED; // IoCompleteRequest()Լ Ͽ,  IRP  ߴ϶ û մϴ
}

void MyISR_DPCRoutine( PKDPC pDpc, PDEVICE_OBJECT DeviceObject, PIRP Irp, PVOID Context )
{
	KIRQL OldIrql;

	DbgPrint("MyISR_DPCRoutine ++ \n" );

	IoAcquireCancelSpinLock( &OldIrql );

	if( DeviceObject->CurrentIrp && (DeviceObject->CurrentIrp->Cancel != TRUE) )
	{
		_SpeakerOff(); // Speaker ϴ
		IoSetCancelRoutine( DeviceObject->CurrentIrp, NULL );
		DeviceObject->CurrentIrp->IoStatus.Status = STATUS_SUCCESS;

		IoReleaseCancelSpinLock( OldIrql );

		IoCompleteRequest( DeviceObject->CurrentIrp, IO_NO_INCREMENT );
		// DeviceObject->CurrentIrp  ISR_DPC_StartIo()Լ ؼ óǰ ִ ֱ IRP Ÿ

		IoStartNextPacket( DeviceObject, TRUE );
		//   IRP ISR_DPC_StartIo()   IO Manager ϴ
	}
	else
		IoReleaseCancelSpinLock( OldIrql );

	DbgPrint("MyISR_DPCRoutine -- \n" );
}

BOOLEAN MyISRRoutine( PKINTERRUPT InterruptObject, PVOID Context )
{
	PDEVICE_OBJECT DeviceObject;

	DeviceObject = (PDEVICE_OBJECT)Context;

	IoRequestDpc( DeviceObject, DeviceObject->CurrentIrp, DeviceObject );

	return TRUE;
}

// Ʒ Լ   DDKԼ ص ϴ
// HalGetBusData( PCIConfiguration, .. );
// ׷,  Լ ObsoleteǾǷ, ڴ Ʒ     Լ Ͽ
// 
void PCIGetBusData( int BusNumber, int DeviceNumber, int FunctionNumber, UCHAR *pDeviceClass, UCHAR *pInterruptLine )
{
	CONFIG_ADDRESS ConfigAddress;
	unsigned char Data[4];
	ULONG *pData;
	unsigned short *pwData;

	*pDeviceClass = 0xFF;
	*pInterruptLine = 0xFF;

	RtlZeroMemory( &ConfigAddress, sizeof(CONFIG_ADDRESS));
	RtlZeroMemory( Data, 4);

	pData = (ULONG *)Data;
	pwData = (unsigned short *)Data;

	ConfigAddress.u.bits.BusNumber = (UCHAR)BusNumber;
	ConfigAddress.u.bits.DeviceNumber = (UCHAR)DeviceNumber;
	ConfigAddress.u.bits.FunctionNumber = (UCHAR)FunctionNumber;
	ConfigAddress.u.bits.RegisterNumber = 0x0; // VendorID, DeviceID

	WRITE_PORT_ULONG( (PULONG)0x0CF8, ConfigAddress.u.AsULONG | 0x80000000 );
	*pData = READ_PORT_ULONG( (PULONG)0x0CFC );

	if( (*pwData == 0) && ( *(pwData+1) == 0 ) )
		return;
	if( (*pwData == 0xFFFF) && ( *(pwData+1) == 0xFFFF ) )
		return;

	DbgPrint("PCI Device Founded!. Vendor ID = 0x%8X, Device ID = 0x%8X\n", *pwData, *(pwData+1) );

	ConfigAddress.u.bits.RegisterNumber = 0x2; // DeviceClass

	WRITE_PORT_ULONG( (PULONG)0x0CF8, ConfigAddress.u.AsULONG | 0x80000000 );
	*pData = READ_PORT_ULONG( (PULONG)0x0CFC );

	*pDeviceClass = Data[3];

	ConfigAddress.u.bits.RegisterNumber = 0x3c/4; // InterruptLine

	WRITE_PORT_ULONG( (PULONG)0x0CF8, ConfigAddress.u.AsULONG | 0x80000000 );
	*pData = READ_PORT_ULONG( (PULONG)0x0CFC );
	
	*pInterruptLine = Data[0];
}

BOOLEAN _registerISR_DPC( PDEVICE_OBJECT DeviceObject )
{
	PDEVICE_EXTENSION deviceExtension;
	int BusNumber;
	int DeviceNumber;
	int FunctionNumber;
	int ret;
	BOOLEAN Found;
	KIRQL Irql;
	KAFFINITY Affinity;
	NTSTATUS status;
	UCHAR InterruptLine;
	UCHAR DeviceClass;
	
	deviceExtension = (PDEVICE_EXTENSION)DeviceObject->DeviceExtension;
	IoInitializeDpcRequest( DeviceObject, MyISR_DPCRoutine );

	// PCIī߿ Networkī带 ãϴ
	Found = FALSE;

#define NETWORK_CLASS 2

	for( BusNumber = 0; BusNumber < 16 ; BusNumber ++ )
	{
		for( DeviceNumber = 0; DeviceNumber < 32 ; DeviceNumber ++ )
		{
			for( FunctionNumber = 0 ; FunctionNumber < 8 ; FunctionNumber++ )
			{
				PCIGetBusData( BusNumber, DeviceNumber, FunctionNumber, &DeviceClass, &InterruptLine );
				if( DeviceClass == 0xFF )
					continue;
				if( DeviceClass != NETWORK_CLASS )
					continue;
				if( InterruptLine == 0 )
					continue;

				Found = TRUE;
				break;
			}
			if( Found == TRUE )
				break;
		}
		if( Found == TRUE )
			break;
	}
	if( Found == FALSE )
		return Found;

	KeInitializeSpinLock( &deviceExtension->SpinLock );
	deviceExtension->InterruptVector = HalGetInterruptVector( PCIBus, BusNumber, InterruptLine, 0, &Irql, &Affinity );

	// APIC ϴ ȯ濡 deviceExtension->InterruptVector  0x100 ū    ִ. 
	// ,  ڴ  ѹƮ ؼ ؾ Ѵ.
	deviceExtension->InterruptVector &= 0x000000FF;

	status = IoConnectInterrupt( &deviceExtension->InterruptObject, MyISRRoutine, DeviceObject, &deviceExtension->SpinLock, 
		deviceExtension->InterruptVector, InterruptLine, InterruptLine, LevelSensitive,
		TRUE, Affinity, FALSE );

	if( NT_SUCCESS( status ) )
		return TRUE;
	else
		return FALSE;
}

NTSTATUS
ISR_DPC_PnpDispatch
	(
	IN PDEVICE_OBJECT DeviceObject,     /* pointer to device object */
	IN PIRP Irp                         /* pointer to an I/O request packet */
	)
{// PNPManager  IRP óϴ ó
	PIO_STACK_LOCATION pStack;                /* pointer to IRP stack */
	PDEVICE_EXTENSION deviceExtension;          /* pointer to device extention */
	NTSTATUS returnStatus = STATUS_SUCCESS;     /* stores status to be returned */
	PDEVICE_OBJECT NextLayerDeviceObject;           /* pointer to the stack device object */
	
	DbgPrint("ISR_DPC_PnpDispatch ++ \n" );

	pStack = IoGetCurrentIrpStackLocation ( Irp );
	deviceExtension = DeviceObject->DeviceExtension;
	NextLayerDeviceObject = deviceExtension->NextLayerDeviceObject;
			
	switch ( pStack->MinorFunction )
	{
		case IRP_MN_START_DEVICE :
		{// DeviceStack ϶ 㰡 ǹմϴ
			KEVENT event;   /* event used to wait for an operation to complete */
			KeInitializeEvent(&event, NotificationEvent, FALSE);
			IoCopyCurrentIrpStackLocationToNext(Irp);  
			IoSetCompletionRoutine(Irp,
								   ISR_DPC_DeferIrpCompletion,
								   &event,
								   TRUE,
								   TRUE,
								   TRUE); // ISR_DPC_DeferIrpCompletion()Լ  IRP  Next Layer Driver ȣϴ
										  // IoCompleteRequest()Լ ؼ ȣǵ մϴ
			
			returnStatus = IoCallDriver(NextLayerDeviceObject,Irp); // NextLayer Driver   ѵ,  Ȯմϴ
			if (returnStatus == STATUS_PENDING) 
			{
				KeWaitForSingleObject(
					&event,
					Executive,
					KernelMode,
					FALSE,
					NULL);
				returnStatus = Irp->IoStatus.Status;
			}
			if( !NT_SUCCESS( returnStatus ) )
			{
				DbgPrint("Start Device Fail\n" );
				IoCompleteRequest( Irp, IO_NO_INCREMENT );
				goto ISR_DPC_PnpDispatch_Exit;
			}
			// NextLayer Driver    ٷ, 츮 ΰ ص ٴ ǹԴϴ

			// Do Job...
			// InterruptServiceRoutine()ϰ DPCƾ մϴ
			if( _registerISR_DPC( DeviceObject ) == FALSE )
			{
				Irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
				IoCompleteRequest( Irp, IO_NO_INCREMENT ); //  IRP ûմϴ
				return STATUS_UNSUCCESSFUL;
			}

			IoSetDeviceInterfaceState( &deviceExtension->UnicodeString, TRUE ); 
			// ڰ 츮 DeviceStack ϴ  

			IoCompleteRequest( Irp, IO_NO_INCREMENT ); //  IRP ûմϴ
			// ISR_DPC_DeferIrpCompletion()Լ 츮 IRP  Next Layer Driver ȣ IoCompleteRequest() 
		}	// ϵ "STATUS_MORE_PROCESSING_REQUIRED" ױ⶧, 츮 û ؾմϴ
		break;

		case IRP_MN_SURPRISE_REMOVAL :
		{ // ڱ ġ PC ŵǾٴ ǹ̿ ޵˴ϴ

			IoSetDeviceInterfaceState( &deviceExtension->UnicodeString, FALSE );
			// ڰ 츮 DeviceStack ϴ  

			Irp->IoStatus.Status = STATUS_SUCCESS;
			IoSkipCurrentIrpStackLocation( Irp ); //  IRP  ɶ ̻   ʰڴٴ ǹԴϴ
			returnStatus = IoCallDriver(NextLayerDeviceObject, Irp);
		}
		break;

		case IRP_MN_REMOVE_DEVICE :
		{ // IRP DeviceStack ü϶ û ǹ̷ ޵˴ϴ
			IoSetDeviceInterfaceState( &deviceExtension->UnicodeString, FALSE );
			// ڰ 츮 DeviceStack ϴ  

			RtlFreeUnicodeString( &deviceExtension->UnicodeString );
			// ڰ ϴ ̸ ޸𸮿 Ѵ

			//InterruptObject õ ۾ ʿ մϴ
			if( deviceExtension->InterruptObject )
			{
				IoDisconnectInterrupt( deviceExtension->InterruptObject );
			}
			//

			IoDetachDevice ( NextLayerDeviceObject );// Next Layer Driver  켱 ϴ
													 // ISR_DPC_AddDevice() IoAttachDeviceToDeviceStack()Լ ݴǹ
			IoDeleteDevice ( DeviceObject );		 // 츮 DeviceObject մϴ
													 // ISR_DPC_AddDevice() IoCreateDevice()Լ ݴǹ
			Irp->IoStatus.Status = STATUS_SUCCESS;
			IoSkipCurrentIrpStackLocation( Irp ); //  IRP  ɶ ̻   ʰڴٴ ǹԴϴ
			returnStatus = IoCallDriver(NextLayerDeviceObject, Irp);
		}
		break;

		default:
		{
			IoSkipCurrentIrpStackLocation( Irp ); //  IRP  ɶ ̻   ʰڴٴ ǹԴϴ
			returnStatus = IoCallDriver(NextLayerDeviceObject, Irp);
		}
		break;
	} 

ISR_DPC_PnpDispatch_Exit:
	DbgPrint("ISR_DPC_PnpDispatch -- \n" );

	return returnStatus;
}

NTSTATUS                                                            
ISR_DPC_PowerDispatch
	(                                                               
		IN PDEVICE_OBJECT DeviceObject,                             
		IN PIRP Irp                                                 
	)
{// POWERManager  IRP óϴ ó
	PDEVICE_EXTENSION		deviceExtension;
	NTSTATUS returnStatus;
	PDEVICE_OBJECT NextLayerDeviceObject;           /* pointer to the stack device object */
	
	DbgPrint("ISR_DPC_PowerDispatch ++ \n" );

	deviceExtension = DeviceObject->DeviceExtension;
	NextLayerDeviceObject = deviceExtension->NextLayerDeviceObject;

	PoStartNextPowerIrp( Irp ); //  ٸ Power IRP   ִٰ PowerManager ˸ϴ
	IoSkipCurrentIrpStackLocation( Irp ); //  IRP  ɶ ̻   ʰڴٴ ǹԴϴ
	returnStatus = PoCallDriver(NextLayerDeviceObject, Irp); // Power IRP ٸ IRP ޸, ݵ PoCallDriver()Լ ؾ մϴ
	
	DbgPrint("ISR_DPC_PowerDispatch -- \n" );
	return returnStatus;
} 

NTSTATUS                                                            
ISR_DPC_Create
	(                                                               
		IN PDEVICE_OBJECT DeviceObject,                             
		IN PIRP Irp                                                 
	)
{// Application  CreateFile()Լ Ǵ ó
	NTSTATUS returnStatus = STATUS_SUCCESS;
	PIO_STACK_LOCATION pStack;
	PDEVICE_EXTENSION deviceExtension;

	DbgPrint("ISR_DPC_Create ++ \n" );

	deviceExtension = (PDEVICE_EXTENSION)DeviceObject->DeviceExtension;

	// ѹ Create , ٸ û  CreateԼ ȣ ϵ ġմϴ
	IoSetDeviceInterfaceState( &deviceExtension->UnicodeString, FALSE );

	DbgPrint("ISR_DPC_Create -- \n" );

	Irp->IoStatus.Status = returnStatus;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );

	return returnStatus;
} 

NTSTATUS                                                            
ISR_DPC_Close
	(                                                               
		IN PDEVICE_OBJECT DeviceObject,                             
		IN PIRP Irp                                                 
	)
{// Application  CloseHandle()Լ Ǵ ó
	NTSTATUS returnStatus = STATUS_SUCCESS;
	PIO_STACK_LOCATION pStack;
	PDEVICE_EXTENSION deviceExtension;

	DbgPrint("ISR_DPC_Close ++ \n" );

	deviceExtension = (PDEVICE_EXTENSION)DeviceObject->DeviceExtension;

	// ѹ Close , ٸ û  CreateԼ ȣ  ֵ ġմϴ
	IoSetDeviceInterfaceState( &deviceExtension->UnicodeString, TRUE );

	DbgPrint("ISR_DPC_Close -- \n" );

	Irp->IoStatus.Status = returnStatus;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );

	return returnStatus;
} 

NTSTATUS                                                            
ISR_DPC_Cleanup
	(                                                               
		IN PDEVICE_OBJECT DeviceObject,                             
		IN PIRP Irp                                                 
	)
{// Application  CloseHandle()Լ Ǵ ó
	NTSTATUS returnStatus = STATUS_SUCCESS;
	PIO_STACK_LOCATION pStack;
	LIST_ENTRY ListHead; // Pendingó  ߰
	KIRQL OldIrql;
	PKDEVICE_QUEUE_ENTRY pTargetDeviceQueueEntry;
	PLIST_ENTRY pTargetListEntry;
	PIRP pTargetIrp;
	PDEVICE_EXTENSION deviceExtension;

	DbgPrint("ISR_DPC_Cleanup ++ \n" );

	deviceExtension = (PDEVICE_EXTENSION)DeviceObject->DeviceExtension;

	InitializeListHead( &ListHead );

	IoAcquireCancelSpinLock( &OldIrql );

	while( pTargetDeviceQueueEntry = KeRemoveDeviceQueue( &(DeviceObject->DeviceQueue) ) )
	{
		pTargetIrp = CONTAINING_RECORD (pTargetDeviceQueueEntry, IRP, Tail.Overlay.DeviceQueueEntry);
		InsertTailList( &ListHead, &(pTargetIrp->Tail.Overlay.ListEntry) );
		IoSetCancelRoutine( pTargetIrp, NULL );
	}

	if( DeviceObject->CurrentIrp )
	{
		IoSetCancelRoutine( DeviceObject->CurrentIrp, NULL );
		KeCancelTimer( &deviceExtension->MyTimer );
		_SpeakerOff();
		
		DeviceObject->CurrentIrp->Cancel = TRUE;

		IoReleaseCancelSpinLock( OldIrql );

		DeviceObject->CurrentIrp->IoStatus.Status = STATUS_CANCELLED;
		IoCompleteRequest( DeviceObject->CurrentIrp, IO_NO_INCREMENT );

		IoStartNextPacket( DeviceObject, TRUE );
	}
	else
		IoReleaseCancelSpinLock( OldIrql );

	while( !IsListEmpty( &ListHead ) )
	{
		pTargetListEntry = RemoveHeadList( &ListHead );
		pTargetIrp = CONTAINING_RECORD (pTargetListEntry, IRP, Tail.Overlay.ListEntry);
		pTargetIrp->IoStatus.Status = STATUS_CANCELLED;
		pTargetIrp->IoStatus.Information = 0;
		IoCompleteRequest( pTargetIrp, IO_NO_INCREMENT );
	}

	DbgPrint("ISR_DPC_Cleanup -- \n" );

	Irp->IoStatus.Status = returnStatus;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );

	return returnStatus;
} 

void CancelRoutine( PDEVICE_OBJECT DeviceObject, PIRP Irp )
{
	PDEVICE_EXTENSION deviceExtension;

	deviceExtension = (PDEVICE_EXTENSION)DeviceObject->DeviceExtension;

	DbgPrint("CancelRoutine ++ \n" );

	if( DeviceObject->CurrentIrp == Irp )
	{
		KeCancelTimer( &deviceExtension->MyTimer );
		_SpeakerOff();

		IoReleaseCancelSpinLock(Irp->CancelIrql); 

		Irp->IoStatus.Status = STATUS_CANCELLED;
		Irp->IoStatus.Information = 0;
		IoCompleteRequest(Irp, IO_NO_INCREMENT); // IRP óûմϴ

		IoStartNextPacket( DeviceObject, TRUE );		
	}
	else
	{
		KeRemoveEntryDeviceQueue( &(DeviceObject->DeviceQueue), &(Irp->Tail.Overlay.DeviceQueueEntry) );
		// CancelSpinLock ݳϱ , 켱 ҵ IRP DeviceQueue ؾ Ѵ

		IoReleaseCancelSpinLock(Irp->CancelIrql); 

		Irp->IoStatus.Status = STATUS_CANCELLED;
		Irp->IoStatus.Information = 0;
		IoCompleteRequest(Irp, IO_NO_INCREMENT); // IRP óûմϴ
	}

	DbgPrint("IRP Cancelled!!\n" );

	DbgPrint("CancelRoutine -- \n" );
}

NTSTATUS
ISR_DPC_DeviceIoControl
	(
	IN PDEVICE_OBJECT DeviceObject,     /* pointer to device object */
	IN PIRP Irp                         /* pointer to an I/O request packet */
	)
{ // α׷ ȣϴ Win32 API DeviceIoControl()Լ Ǵ Լ
	NTSTATUS returnStatus = STATUS_SUCCESS;
	PIO_STACK_LOCATION pStack;
	ULONG dwRequestSize=0;

	DbgPrint("ISR_DPC_DeviceIoControl ++ \n" );

	pStack = IoGetCurrentIrpStackLocation( Irp );

	switch( pStack->Parameters.DeviceIoControl.IoControlCode )
	// α׷ ϴ ControlCode Ȯմϴ. 
	// ̰ Ķ ȿ Ȯմϴ.
	{
		case IOCTL_SAMPLE_SET_FREQUENCY:
			returnStatus = STATUS_PENDING;
			break;
		default:
			returnStatus = STATUS_INVALID_PARAMETER;
			break;
	}

	if( returnStatus == STATUS_PENDING )
	{
		IoMarkIrpPending( Irp ); // Irp 񵿱 Ұ ŷ
		IoStartPacket( DeviceObject, Irp, NULL, CancelRoutine ); // CancelRoutine ϵ
		// Irp IO Manager ñ
		// ISR_DPC_StartIo()Լ IO Manager ؼ Ǿ 䱸
	}
	else
	{
		Irp->IoStatus.Information = 0;
		Irp->IoStatus.Status = returnStatus;
		IoCompleteRequest( Irp, IO_NO_INCREMENT );
	}

	DbgPrint("ISR_DPC_DeviceIoControl -- \n" );

	return returnStatus;
}

void
ISR_DPC_StartIo
	(
	IN PDEVICE_OBJECT DeviceObject,     /* pointer to device object */
	IN PIRP Irp                         /* pointer to an I/O request packet */
	)
{
	NTSTATUS returnStatus = STATUS_SUCCESS;
	PIO_STACK_LOCATION pStack;
	ULONG dwRequestSize=0;
	PVOID pBuffer;
	PDEVICE_EXTENSION deviceExtension;
	KIRQL OldIrql;

	DbgPrint("ISR_DPC_StartIo ++ \n" );

	deviceExtension = (PDEVICE_EXTENSION)DeviceObject->DeviceExtension;

	IoAcquireCancelSpinLock( &OldIrql );

	if( ( DeviceObject->CurrentIrp != Irp ) || ( Irp->Cancel ) )
	{
		IoReleaseCancelSpinLock( OldIrql );
		return;
	}

	pStack = IoGetCurrentIrpStackLocation( Irp );

	switch( pStack->Parameters.DeviceIoControl.IoControlCode )
	// α׷ ϴ ControlCode Ȯմϴ. 
	// ̰ Ķ ȿ Ȯմϴ.
	{
		case IOCTL_SAMPLE_SET_FREQUENCY:
			// Ŀ Ѱ ļ մϴ
			// ļ α׷ ݴϴ
			pBuffer = Irp->AssociatedIrp.SystemBuffer;
			_SpeakerOn( *(unsigned short *)pBuffer );

			// Timer  ϴ ϵ̹Ƿ, Ÿ̸(2) ŵϴ
#define TIME_2S 20000000 //  100ns

			KeSetTimer( &deviceExtension->MyTimer, 
				RtlConvertLongToLargeInteger( 20000000 * -1 ), 
				&deviceExtension->MyDpc );

			break;
	}

	DbgPrint("ISR_DPC_StartIo -- \n" );

	IoReleaseCancelSpinLock( OldIrql );
}

unsigned char InterruptCallFrame[3];

typedef
PVOID  
(*PINTERRUPTCALL) (
    );

VOID
MyDpcRoutine(
    IN PKDPC Dpc,
    IN PVOID DeferredContext,
    IN PVOID SystemArgument1,
    IN PVOID SystemArgument2
    )
{
	PDEVICE_OBJECT DeviceObject;
	PDEVICE_EXTENSION deviceExtension;
	PINTERRUPTCALL pInterruptCall;

	DeviceObject = (PDEVICE_OBJECT)DeferredContext;
	deviceExtension = (PDEVICE_EXTENSION)DeviceObject->DeviceExtension;

	DbgPrint("MyDpcRoutine ++ \n" );
	
	//  ϵ ͷƮ ߻Ѱó 䳻  ؼ ƮͷƮ ߻ŵϴ
	InterruptCallFrame[0] = 0xCD; // INT xx
	InterruptCallFrame[1] = (unsigned char)deviceExtension->InterruptVector;
	InterruptCallFrame[2] = 0xC3; // RET

	pInterruptCall = (PINTERRUPTCALL)InterruptCallFrame;
	pInterruptCall();

	DbgPrint("MyDpcRoutine -- \n" );
}
